use std::str::FromStr;

use palette::{
    color_difference::Wcag21RelativeContrast, color_theory::*, convert::FromColorUnclamped, Darken,
    FromColor, Hsl, IsWithinBounds, Lab, Lighten, Mix, Oklch, ShiftHue, Srgb,
};
use wasm_bindgen::prelude::*;

mod errors;

#[wasm_bindgen]
pub fn represent_rgb(color: &str) -> Result<Box<[u8]>, errors::ColorError> {
    let srgb = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<u8>();
    Ok(Box::new([srgb.red, srgb.green, srgb.blue]))
}

#[wasm_bindgen]
pub fn rgb_to_hex(r: u8, g: u8, b: u8) -> Result<String, errors::ColorError> {
    let srgb = Srgb::new(r, g, b);
    Ok(format!("{:x}", srgb))
}

#[wasm_bindgen]
pub fn represent_hsl(color: &str) -> Result<Box<[f32]>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let hsl = Hsl::from_color(origin_color);
    Ok(Box::new([
        hsl.hue.into_positive_degrees(),
        hsl.saturation,
        hsl.lightness,
    ]))
}

#[wasm_bindgen]
pub fn hsl_to_hex(h: f32, s: f32, l: f32) -> Result<String, errors::ColorError> {
    let hsl = Hsl::new(h, s, l);
    let srgb = Srgb::from_color_unclamped(hsl);
    if !srgb.is_within_bounds() {
        return Err(errors::ColorError::ComponentOutOfBounds);
    }
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn represent_lab(color: &str) -> Result<Box<[f32]>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let lab = Lab::from_color(origin_color);
    Ok(Box::new([lab.l, lab.a, lab.b]))
}

#[wasm_bindgen]
pub fn lab_to_hex(l: f32, a: f32, b: f32) -> Result<String, errors::ColorError> {
    let lab = Lab::new(l, a, b);
    let srgb = Srgb::from_color_unclamped(lab);
    if !srgb.is_within_bounds() {
        return Err(errors::ColorError::ComponentOutOfBounds);
    }
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn represent_oklch(color: &str) -> Result<Box<[f32]>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let lch = Oklch::from_color(origin_color);
    Ok(Box::new([
        lch.l,
        lch.chroma,
        lch.hue.into_positive_degrees(),
    ]))
}

#[wasm_bindgen]
pub fn oklch_to_hex(l: f32, c: f32, h: f32) -> Result<String, errors::ColorError> {
    let lch = Oklch::new(l, c, h);
    let srgb = Srgb::from_color_unclamped(lch);
    if !srgb.is_within_bounds() {
        return Err(errors::ColorError::ComponentOutOfBounds);
    }
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn shift_hue(color: &str, degree: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let shifted_color = oklch.shift_hue(degree);
    let srgb = Srgb::from_color(shifted_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn lighten(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let lightened_color = oklch.lighten(percent);
    let srgb = Srgb::from_color(lightened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn lighten_absolute(color: &str, value: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let lightened_color = oklch.lighten_fixed(value);
    let srgb = Srgb::from_color(lightened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn darken(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let darkened_color = oklch.darken(percent);
    let srgb = Srgb::from_color(darkened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn darken_absolute(color: &str, value: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let darkened_color = oklch.darken_fixed(value);
    let srgb = Srgb::from_color(darkened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn mix(color1: &str, color2: &str, percent: f32) -> Result<String, errors::ColorError> {
    let srgb1 = Srgb::from_str(color1)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color1.to_string()))?
        .into_format::<f32>();
    let srgb2 = Srgb::from_str(color2)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color2.to_string()))?
        .into_format::<f32>();
    let mixed_color = srgb1.mix(srgb2, percent);
    Ok(format!("{:x}", mixed_color.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn tint(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let tinted_color = origin_color.mix(Srgb::new(1.0, 1.0, 1.0), percent);
    Ok(format!("{:x}", tinted_color.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn shade(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let shaded_color = origin_color.mix(Srgb::new(0.0, 0.0, 0.0), percent);
    Ok(format!("{:x}", shaded_color.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn wacg_relative_contrast(fg_color: &str, bg_color: &str) -> Result<f32, errors::ColorError> {
    let fg_srgb = Srgb::from_str(fg_color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(fg_color.to_string()))?
        .into_format::<f32>();
    let bg_srgb = Srgb::from_str(bg_color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(bg_color.to_string()))?
        .into_format::<f32>();
    Ok(fg_srgb.relative_contrast(bg_srgb))
}

#[wasm_bindgen]
pub fn analogous_30(color: &str) -> Result<Vec<String>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let (color_n30, color_p30) = oklch.analogous();
    let srgb_n30 = Srgb::from_color(color_n30);
    let srgb_p30 = Srgb::from_color(color_p30);
    Ok(vec![
        format!("{:x}", srgb_n30.into_format::<u8>()),
        format!("{:x}", srgb_p30.into_format::<u8>()),
    ])
}

#[wasm_bindgen]
pub fn analogous_60(color: &str) -> Result<Vec<String>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let (color_n60, color_p60) = oklch.analogous_secondary();
    let srgb_n60 = Srgb::from_color(color_n60);
    let srgb_p60 = Srgb::from_color(color_p60);
    Ok(vec![
        format!("{:x}", srgb_n60.into_format::<u8>()),
        format!("{:x}", srgb_p60.into_format::<u8>()),
    ])
}

#[wasm_bindgen]
pub fn complementary(color: &str) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let complementary_color = oklch.complementary();
    let srgb = Srgb::from_color(complementary_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn split_complementary(color: &str) -> Result<Vec<String>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let (color_p150, color_p210) = oklch.split_complementary();
    let srgb_p150 = Srgb::from_color(color_p150);
    let srgb_p210 = Srgb::from_color(color_p210);
    Ok(vec![
        format!("{:x}", srgb_p150.into_format::<u8>()),
        format!("{:x}", srgb_p210.into_format::<u8>()),
    ])
}

#[wasm_bindgen]
pub fn tetradic(color: &str) -> Result<Vec<String>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let (color_p90, color_p180, color_p270) = oklch.tetradic();
    let srgb_p90 = Srgb::from_color(color_p90);
    let srgb_p180 = Srgb::from_color(color_p180);
    let srgb_p270 = Srgb::from_color(color_p270);
    Ok(vec![
        format!("{:x}", srgb_p90.into_format::<u8>()),
        format!("{:x}", srgb_p180.into_format::<u8>()),
        format!("{:x}", srgb_p270.into_format::<u8>()),
    ])
}

#[wasm_bindgen]
pub fn triadic(color: &str) -> Result<Vec<String>, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let (color_p120, color_p240) = oklch.triadic();
    let srgb_p120 = Srgb::from_color(color_p120);
    let srgb_p240 = Srgb::from_color(color_p240);
    Ok(vec![
        format!("{:x}", srgb_p120.into_format::<u8>()),
        format!("{:x}", srgb_p240.into_format::<u8>()),
    ])
}
